//---------------------------------------------------------------------------
// Sound Blaster PID regulator experiment - model poles/zeros viewer
//---------------------------------------------------------------------------
// Experimental Sound Blaster PID regulator. It uses SB as an input/output
// interface to the process to be controlled. This demo is designed as
// a speed loop for DC motors with with PWM control and rotary encoder
// feedback but it can be easily converted to any other kind of input/ouput
// format.
// Regulator uses beta-I-PD topology for a controller, variable step
// Steepest-Descent algorithm for system identification and Ziegler-Nichols
// method for auto-tuning.
// For more details follow function comments and visit project www.
//
// License:
// This code can be used in any project but it was designed only for learning
// purposes so author is not responsible for its proper function!
//
// (c) 2012, Stanislav Maslan, s.maslan@seznam.cz
// http://elektronika.kvalitne.cz/SW/SB_regulator.html
//---------------------------------------------------------------------------
#include <vcl.h>
#include <math.h>
#pragma hdrstop

#include "MPZ_form.h"
#include "graph.h"
#include "sys_model.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMPZForm *MPZForm;
//---------------------------------------------------------------------------
__fastcall TMPZForm::TMPZForm(TComponent* Owner)
	: TForm(Owner)
{
	DoubleBuffered=true;
}
//---------------------------------------------------------------------------
void __fastcall TMPZForm::FormResize(TObject *Sender)
{
	DrawPZ();
}
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// redraw P/Z map
//---------------------------------------------------------------------------
void TMPZForm::DrawPZ(void)
{
	int i;

	// get ARX model zeros/poles
	TPoZe pole[3];
	int pn=3;
	TPoZe zero[3];
	int zn=3;
	sys_idn_get_PZ(pole,&pn,zero,&zn);

	// window size
	int xs=img->Width;
	int ys=img->Height;

	// find scale
	double mx=1.0;
	double my=1.0;
	for(i=0;i<pn;i++)
	{
		// limit max size just in case
		if(fabs(pole[i].re)>100 || fabs(pole[i].im)>100)
			continue;

		if(fabs(pole[i].re)>mx)
			mx=fabs(pole[i].re);
		if(fabs(pole[i].im)>my)
			my=fabs(pole[i].im);
	}
	for(i=0;i<zn;i++)
	{
		// limit max size just in case
		if(fabs(zero[i].re)>100 || fabs(zero[i].im)>100)
			continue;

		if(fabs(zero[i].re)>mx)
			mx=fabs(zero[i].re);
		if(fabs(zero[i].im)>my)
			my=fabs(zero[i].im);
	}
	// extend
	mx*=PZMAP_SL;
	my*=PZMAP_SL;

	// scale
	double ss=((double)xs/mx>(double)ys/my)?(0.5*(double)ys/my):(0.5*(double)xs/mx);

  // img font
	img->Picture->Bitmap->Canvas->Font=FD->Font;

	// draw bakcground
	PZ_draw_back(img,ss);

	// draw poles
	PZ_draw_pz(img,pole,pn,PZMAP_TYPE_POLE,ss);

	// draw zeros
	PZ_draw_pz(img,zero,zn,PZMAP_TYPE_ZERO,ss);

	// repaint
	img->Repaint();
}

void __fastcall TMPZForm::imgDblClick(TObject *Sender)
{
	if(OD1->Execute())
	{
		if(strcmpi(ExtractFileExt(OD1->FileName).c_str(),".bmp")!=0)
			OD1->FileName=OD1->FileName+".bmp";
		img->Picture->Bitmap->PixelFormat=pf24bit;
		img->Picture->Bitmap->SaveToFile(OD1->FileName);
	}
}
//---------------------------------------------------------------------------

