//---------------------------------------------------------------------------
#include <math.h>
#include <stdlib.h>
#include <string.h>

#pragma hdrstop

#include "sys_model.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)


TSysId sys;

void sys_idn_init(void)
{
	sys.x=NULL;
	sys.y=NULL;
	sys.len=0;
	sys.pos=0;
	sys.n=0;
	sys.filter=1;

	for(int i=0;i<SYS_N/2;i++)
		sys.yp[i]=0.0;
}

void sys_idn_close(void)
{
	if(sys.x)
		free((void*)sys.x);
	if(sys.y)
		free((void*)sys.y);
	sys_idn_init();
}

int sys_idn_config(int len,double Ts)
{
	if(len<12)
		len=12;

	sys.Ts=Ts;

	if(len!=sys.len)
	{
		// loose old buffers
		if(sys.x)
			free((void*)sys.x);
		if(sys.y)
			free((void*)sys.y);
		sys_idn_init();

		// allocate buffers
		sys.x=(double*)malloc(len*sizeof(double));
		if(!sys.x)
		{
			sys_idn_close();
			return(1);
		}
		sys.y=(double*)malloc(len*sizeof(double));
		if(!sys.y)
		{
			sys_idn_close();
			return(1);
		}

		// set new len
		sys.len=len;
	}

	return(0);
}

void sys_idn_filter(int len)
{
	sys.filter=len;
}

int sys_idn_update(double nx,double ny,double *ypred,double *theta,double *kp)
{
	int i,j,k;

	if(!sys.len)
		return(1);

	// add new valus into buffers
	sys.x[sys.pos]=nx;
	sys.y[sys.pos]=ny;

	// update round buffer position
	sys.pos++;
	if(sys.pos>=sys.len)
		sys.pos=0;

	// update sample count
	sys.n++;
	if(sys.n>sys.len)
		sys.n=sys.len;

	// allocate work buffers
	double *x=(double*)malloc(sys.n*sizeof(double));
	if(!x)
	{
		return(1);
	}
	double *y=(double*)malloc(sys.n*sizeof(double));
	if(!y)
	{
		free((void*)x);
		return(1);
	}

	// fill them with all available data
	int pos=sys.pos;
	for(i=0;i<sys.n;i++)
	{
		pos--;
		if(pos<0)
			pos=sys.n-1;

		x[i]=sys.x[pos];
		y[i]=sys.y[pos];
	}

	// temp matrix
	double A[SYS_N][SYS_N*2];

	// === PHI matrix top row (third order ARX model) ===
	// y[i] = b1*x[i-1] + b2*x[i-2] + b3*x[i-3] - a1*y[i-1] - a2*y[i-2] - a3*y[i-3];
	double *pPHI[SYS_N]={&x[1],&x[2],&x[3],&y[1],&y[2],&y[3]};
	double th[SYS_N];
	int n=sys.n-4;

	// === start least square method ===
	// A = PHI'*PHI
	for(i=0;i<SYS_N;i++)
	{
		for(j=0;j<SYS_N;j++)
		{
			double *row=pPHI[i];
			double *col=pPHI[j];

			// sum(row[]*col[])
			double sum=0.0;
			for(k=0;k<n;k++)
				sum+=row[k]*col[k];

			// store sum into resulting matrix
			A[i][j]=sum;
		}
	}

	// === inverse matrix A: (PHI'*PHI)^-1 ===
	// A = [A eye(SYS_N)]
	for(i=0;i<SYS_N;i++)
		for(j=0;j<SYS_N;j++)
			A[i][j+SYS_N]=(i==j)?1.0:0.0;

	// now gauss elim bottom echelon
	for(i=0;i<SYS_N;i++)
	{
		// scale actual column top item to 1.0
		if(!A[i][i])
			goto sys_idn_sing;
		double c=1/A[i][i];
		for(k=0;k<SYS_N*2;k++)
			A[i][k]*=c;

		if(i==SYS_N-1)
			break;

		// for each minor row
		for(j=i+1;j<SYS_N;j++)
		{
			// eliminate column
			if(!A[i][i])
				goto sys_idn_sing;
			c=A[j][i];
			for(k=i;k<SYS_N*2;k++)
				A[j][k]-=c*A[i][k];
			A[j][i]=0.0;
		}
	}
	// now gauss elim top echelon
	for(i=1;i<SYS_N;i++)
	{
		// for each minor row
		for(j=0;j<i;j++)
		{
			// eliminate column
			double c=A[j][i];
			for(k=i;k<SYS_N*2;k++)
				A[j][k]-=c*A[i][k];
		}
	}
	// now hopefully right half of A should contain inversion (if it wasn't singular) :)

	// === now finish least square method ===
	// theta = (PHI'*PHI)^-1*PHI'*Y
	for(i=0;i<SYS_N;i++)
	{
		double tht=0.0;
		for(j=0;j<n;j++)
		{
			double sum=0.0;
			for(k=0;k<SYS_N;k++)
				sum+=A[i][k+SYS_N]*pPHI[k][j];

			// theta(i) += sum(A(i,:).*PHI(:,k),k=0..n-4)
			tht+=sum*y[j];
		}
		// store final model parameter
		th[i]=(i>=SYS_N/2)?(-tht):tht;
	}

	// filter identified ARX model parameters
	for(i=0;i<SYS_N;i++)
		sys.th[i]=(sys.th[i]*(double)(sys.filter-1) + th[i])/(double)sys.filter;

	// in case matrix was singular
	sys_idn_sing:

	// predict system output (ARX model output)
	double y_pred=0.0;
	for(i=0;i<SYS_N/2;i++)
		y_pred += x[i+1]*sys.th[i] - sys.yp[i]*sys.th[i+SYS_N/2];

	if(y_pred>1000)
		y_pred=1000;
	else if(y_pred<-1000)
		y_pred=-1000;

	// update output memory
	for(i=SYS_N/2-1;i>0;i--)
		sys.yp[i]=sys.yp[i-1];
	sys.yp[0]=y_pred;


	// and return if requested
	if(ypred)
		*ypred=y_pred;

	// loose temp buffers
	free((void*)x);
	free((void*)y);

	// return ARX parameters if requested
	if(theta)
		memcpy((void*)theta,(void*)sys.th,SYS_N*sizeof(double));


	// === Now the fun beggins: calculation of critical parameters for 3rd order ARX :) ===
	double b1=sys.th[0];
	double b2=sys.th[1];
	double b3=sys.th[2];
	double a1=sys.th[3];
	double a2=sys.th[4];
	double a3=sys.th[5];

	if(b1!=0.0 && b2!=0.0 && b3!=0.0 && a1!=0.0 && a2!=0.0 && a3!=0.0)
	{
		double r0=a3*(a3-a1)+a2-1;
		double r1=b3*(2*a3-a1)+b2-a3*b1;
		double r2=b3*(b3-b1);
		double d=r1*r1-4*r0*r2;
		if(d<0)
			d=0;

		double Kp1=(-r1+sqrt(d))/2/r2;
		double Kp2=(-r1-sqrt(d))/2/r2;
		double Kp3=(1-a1+a2-a3)/(b1-b2+b3);
		double e11=a3+b3*Kp1;
		double e12=a3+b3*Kp2;

		double Tk=0.0;
		double Kpk;
		if(Kp1>0 && fabs(e11)<1)
				Kpk=Kp1;
		else if(Kp2>0 && fabs(e12)<1)
				Kpk=Kp2;
		else
		{
				Kpk=Kp3;
				Tk=2*sys.Ts;
		}

		if(Tk==0.0)
		{
				double al=(a3-a1+Kpk*(b3-b1))/2;
				if(al>1.0)
					al=1.0;
				else if(al<-1.0)
					al=-1.0;

				double wk=1/sys.Ts*acos(al);
				if(wk==0)
					wk=1E-9;
				Tk=2*3.1415/wk;
		}

		// and here we go: critical gain and oscilation frequency
		Kpk=fabs(Kpk);
		Tk=fabs(Tk);

		sys.kp[0]=Kpk;
		sys.kp[1]=Tk;

		// return critical parameters if requested
		if(kp)
			memcpy((void*)kp,(void*)sys.kp,2*sizeof(double));

		/*reg.Ti=1.0*Tk;
		reg.Td=0.125*Tk;
		reg.k=0.3*Kpk;*/
	}


	return(0);
}